# -*- coding:utf-8 -*-

import os
import urllib2

import caldav
import ics

import time
from datetime import datetime, timedelta

# disable warnings
import requests
requests.packages.urllib3.disable_warnings()


class SyncToCalDav(object):
	def __init__(self, ics_url=None, url=None, username=None, password=None, calendar_id=None, purge_calendar=True, delay_purge=False):
		self.ics_url = ics_url
		self.url = url
		self.username = username
		self.password = password
		self.calendar_id = calendar_id
		self.purge_calendar = purge_calendar
		self.delay_purge = delay_purge
		self.purge_uids = []
		self.calendar = None
		self.client = caldav.DAVClient(url=self.url, username=self.username, password=self.password, ssl_verify_cert=False)
		self.needs_normalization = False

	def parse_and_upload_data(self, ical_data, needs_normalization=False):
		self.needs_normalization = needs_normalization
		events = self.split_events(ical_data)
		self.upload_to_dav(events)

	def get_vcal_content(self):
		if os.path.exists(self.ics_url):
			with open(self.ics_url) as f:
				return f.read()
		else:
			return urllib2.urlopen(self.ics_url).read()

	def split_events(self, vcal):
		if isinstance(vcal, str):
			c = ics.Calendar(vcal.decode('utf-8'))
		else:
			c = ics.Calendar(vcal)
		return c.events

	def connect_to_dav_calendar(self):
		client = caldav.DAVClient(url=self.url, username=self.username, password=self.password, ssl_verify_cert=False)
		# fix for apache mod_rewrite
		client.headers["User-Agent"] = "Calendar"
		principal = client.principal()
		calendars = principal.calendars()
		# select calender to work with
		for cal in calendars:
			cu = cal.canonical_url
			if self.calendar_id and cu.endswith('/%s/' % self.calendar_id):
				self.calendar = cal
		if not self.calendar_id:
			self.calendar = calendars[0]
		if self.calendar is None:
			if self.calendar_id:
				raise Exception('no calendar found with id `%s`', self.calendar_id)
			else:
				raise Exception('no calendar found.')
			return

	def upload_to_dav(self, events):
		self.connect_to_dav_calendar()

		# purge calendar if needed
		self.purge()

		# import event to calendar
		for event in events:
			if not event:
				continue
			ev = self.stringify_event(event)
			self.add_event_to_calendar(ev)
		# now purge them
		self.delayed_purge()

	def stringify_event(self, event):
		if self.needs_normalization:
			clean_event = self.compensate_timezone(event)
		else:
			clean_event = str(event)
		ev = (
			'BEGIN:VCALENDAR\n'
			'PRODID:Open-Xchange\n'
			'VERSION:2.0\n'
			'CALSCALE:GREGORIAN\n'
			'METHOD:PUBLISH\n'
			'BEGIN:VTIMEZONE\n'
			'TZID:Europe/Berlin\n'
			'TZURL:http://tzurl.org/zoneinfo/Europe/Berlin\n'
			'X-LIC-LOCATION:Europe/Berlin\n'
			'BEGIN:DAYLIGHT\n'
			'TZOFFSETFROM:+0100\n'
			'TZOFFSETTO:+0200\n'
			'TZNAME:CEST\n'
			'DTSTART:19810329T020000\n'
			'RRULE:FREQ=YEARLY;BYMONTH=3;BYDAY=-1SU\n'
			'END:DAYLIGHT\n'
			'BEGIN:STANDARD\n'
			'TZOFFSETFROM:+0200\n'
			'TZOFFSETTO:+0100\n'
			'TZNAME:CET\n'
			'DTSTART:19961027T030000\n'
			'RRULE:FREQ=YEARLY;BYMONTH=10;BYDAY=-1SU\n'
			'END:STANDARD\n'
			'BEGIN:STANDARD\n'
			'TZOFFSETFROM:+005328\n'
			'TZOFFSETTO:+0100\n'
			'TZNAME:CET\n'
			'DTSTART:18930401T000000\n'
			'RDATE:18930401T000000\n'
			'END:STANDARD\n'
			'BEGIN:DAYLIGHT\n'
			'TZOFFSETFROM:+0100\n'
			'TZOFFSETTO:+0200\n'
			'TZNAME:CEST\n'
			'DTSTART:19160430T230000\n'
			'RDATE:19160430T230000\n'
			'RDATE:19170416T020000\n'
			'RDATE:19180415T020000\n'
			'RDATE:19400401T020000\n'
			'RDATE:19430329T020000\n'
			'RDATE:19440403T020000\n'
			'RDATE:19450402T020000\n'
			'RDATE:19460414T020000\n'
			'RDATE:19470406T030000\n'
			'RDATE:19480418T020000\n'
			'RDATE:19490410T020000\n'
			'RDATE:19800406T020000\n'
			'END:DAYLIGHT\n'
			'BEGIN:STANDARD\n'
			'TZOFFSETFROM:+0200\n'
			'TZOFFSETTO:+0100\n'
			'TZNAME:CET\n'
			'DTSTART:19161001T010000\n'
			'RDATE:19161001T010000\n'
			'RDATE:19170917T030000\n'
			'RDATE:19180916T030000\n'
			'RDATE:19421102T030000\n'
			'RDATE:19431004T030000\n'
			'RDATE:19441002T030000\n'
			'RDATE:19451118T030000\n'
			'RDATE:19461007T030000\n'
			'RDATE:19471005T030000\n'
			'RDATE:19481003T030000\n'
			'RDATE:19491002T030000\n'
			'RDATE:19800928T030000\n'
			'RDATE:19810927T030000\n'
			'RDATE:19820926T030000\n'
			'RDATE:19830925T030000\n'
			'RDATE:19840930T030000\n'
			'RDATE:19850929T030000\n'
			'RDATE:19860928T030000\n'
			'RDATE:19870927T030000\n'
			'RDATE:19880925T030000\n'
			'RDATE:19890924T030000\n'
			'RDATE:19900930T030000\n'
			'RDATE:19910929T030000\n'
			'RDATE:19920927T030000\n'
			'RDATE:19930926T030000\n'
			'RDATE:19940925T030000\n'
			'RDATE:19950924T030000\n'
			'END:STANDARD\n'
			'BEGIN:DAYLIGHT\n'
			'TZOFFSETFROM:+0200\n'
			'TZOFFSETTO:+0300\n'
			'TZNAME:CEMT\n'
			'DTSTART:19450524T020000\n'
			'RDATE:19450524T020000\n'
			'RDATE:19470511T030000\n'
			'END:DAYLIGHT\n'
			'BEGIN:DAYLIGHT\n'
			'TZOFFSETFROM:+0300\n'
			'TZOFFSETTO:+0200\n'
			'TZNAME:CEST\n'
			'DTSTART:19450924T030000\n'
			'RDATE:19450924T030000\n'
			'RDATE:19470629T030000\n'
			'END:DAYLIGHT\n'
			'BEGIN:STANDARD\n'
			'TZOFFSETFROM:+0100\n'
			'TZOFFSETTO:+0100\n'
			'TZNAME:CET\n'
			'DTSTART:19460101T000000\n'
			'RDATE:19460101T000000\n'
			'RDATE:19800101T000000\n'
			'END:STANDARD\n'
			'END:VTIMEZONE\n'
			'%s\n'
			'END:VCALENDAR'
			) % clean_event
		return ev

	def add_event_to_calendar(self, event):
		try:
			self.calendar.add_event(event)
		except(caldav.lib.error.PutError):
			# try again
			# wait 5 minutes for ox to recover
			print("Sleeping for 5 minutes")
			time.sleep(300)
			self.calendar.add_event(event)

	def compensate_timezone(self, event):
		# workaround for ical module adding Zulu timezone
		# to every calendar entry on casting to string

		split_event = str(event).split('\n')
		clean_split_event = []
		for event_part in split_event:
			if event_part.startswith('DT') and event_part.endswith('Z'):
				_parts = event_part.split(':')
				normalized_date = self.normalize_date_time(_parts[1])
				event_part = '%s;TZID=Europe/Berlin:%s' % (_parts[0], normalized_date,)
			clean_split_event.append(event_part)
		cleaned_event = '\n'.join(clean_split_event)
		return cleaned_event

	def normalize_date_time(self, arrowstring):
		arrow_format = "%Y%m%dT%H%M%SZ"
		new_arrow_format = "%Y%m%dT%H%M%S"
		date = datetime.strptime(arrowstring, arrow_format)
		offset = timedelta(hours=1)
		date_new = date + offset
		new_arrowstring = datetime.strftime(date_new, new_arrow_format)
		return new_arrowstring

	def delayed_purge(self):
		for uid in self.purge_uids:
			ev = self.calendar.event_by_uid(uid)
			ev.delete()

	def purge(self):
		l = len(self.calendar.events())
		if self.purge_calendar and l > 0:
			if not self.delay_purge:
				for ev in self.calendar.events():
					if self.delay_purge:
						uid = ev.instance.vevent.uid.value
						self.purge_uids.append(uid)
					else:
						ev.delete()
