#!/usr/bin/python2.7
# -*- coding:utf-8 -*-
#
#
#
#
from sshtunnel import SSHTunnelForwarder
import MySQLdb
import datetime
import SyncToCalDav
import ConfigParser
import optparse
import os
import re
from ics import Calendar, Event

## GLOBALS ##
### Path to the credential files (home path) and the ID of the OX calender
### in which the generated calendar will be exported
OX_CONFIG_FILE = os.path.join(os.getenv('HOME'), '.resource_calendar.secret')
CALDAV_CREDENTIALS_FILE = os.path.join(os.getenv('HOME'), '.calendar.secret')
CALDAV_CALENDAR_ID = '1351'
ROOMS = {"Besprechungsraum": "Bespr. Raum"}
ROOMS_PATTERN = re.compile(r'\b(' + '|'.join(ROOMS.keys()) + r')\b')
DEBUG = 0

### Build up an SSH tunnel to reach the SQL database on the OX server
class Tunnel:
	## 
	def __init__(self, config):
		self.remote_host = "127.0.0.1"
		self.remote_port = 3306
		self.local_port = 3306
		self.ssh_host = config.get('RESOURCE_CALENDAR', 'OX_SERVER_IP')
		self.ssh_port = 22
		self.ssh_user = config.get('RESOURCE_CALENDAR', 'OX_SERVER_SSH_USER')
		self.ssh_key = config.get('RESOURCE_CALENDAR', 'OX_SERVER_SSH_KEY')
		self.server = SSHTunnelForwarder(
			(self.ssh_host, self.ssh_port),
			ssh_username=self.ssh_user,
			ssh_pkey=self.ssh_key,
			remote_bind_address=(self.remote_host, self.remote_port))
		self.port = None

	## 
	def connect(self):
		self.server.start()
		self.port = self.server.local_bind_port

	## 
	def disconnect(self):
		self.server.stop()
		self.port = None

### Query the OX SQL database and return the result
class Mysql:
	## 
	def __init__(self, port, config):
		global DEBUG

		self.query = """
		SELECT dates.timestampfield01 AS Start, dates.timestampfield02 AS Ende, resource.displayName, dates.organizer, dates.intfield07 AS allday, dates.field06 AS misc, dates.field07 AS exception1, dates.field08 AS exception2
		FROM resource
		LEFT JOIN prg_date_rights as dr on resource.id = dr.id
		LEFT JOIN prg_dates as dates on dates.intfield01 = dr.object_id
		WHERE 
		dates.timestampfield01 <= date_add(now(), interval 8 month) && dates.timestampfield02 >= date_sub(now(), interval 7 day)
		;
		"""

		self.port = port
		self.password = config.get('RESOURCE_CALENDAR', 'OX_DB_PASSWORD')
		self.user = config.get('RESOURCE_CALENDAR', 'OX_DB_USER')
		self.dbname = config.get('RESOURCE_CALENDAR', 'OX_DB_NAME')

	## 
	def do_query(self):
		con = MySQLdb.connect(host='127.0.0.1', port=self.port, user=self.user, passwd=self.password, db=self.dbname)
		cur = con.cursor()
		cur.execute(self.query)
		result = cur.fetchall()
		return result

## 
class Ical:
	## 
	def __init__(self, entries, config):
		self.entries = entries
		self.calendar = Calendar()

		# Limit for open ended event series and lower limit to only generate events one week ago
		now = datetime.datetime.now()
		self.limit = now + datetime.timedelta(days=8*365/12)
		self.lower_limit = now - datetime.timedelta(days=7)

		#
		self.dav = None
		self.init_dav(config)

	## 
	def init_dav(self, config):
		url = config.get('CALDAV', 'URL')
		password = config.get('CALDAV', 'PASSWORD')
		user = config.get('CALDAV', 'USER')
		self.dav = SyncToCalDav.SyncToCalDav(url=url, password=password, username=user, calendar_id=CALDAV_CALENDAR_ID)

	## 
	def build_ical_structure(self):
		global DEBUG, ROOMS, ROOMS_PATTERN
		# For each event with an assigned resource
		counter = 0
		for line in self.entries:
			if(DEBUG >= 2): 
				print("Line: " + str(counter))

			_allDay = None
			_start = line[0]
			_end_limit = _end = line[1]
			_room = ROOMS_PATTERN.sub(lambda x: ROOMS[x.group()], line[2])
			_organizer = line[3].split('@')[0]
			name = "%s - %s" % (_room, _organizer,)
			if len(line) > 4:
				_allDay = line[4]
			days = weeks = months = False
			limit_start = limit_end = None
			excludes = []

			#
			if(_allDay):
				for i in range((_end.date() - _start.date()).days):
					day = _start + datetime.timedelta(days=i)
					if(DEBUG >= 2): 
						print("allday",self.dates_to_str(day, prefix="Day: ", allday=True), name, self.dates_to_str(excludes, allday=True, prefix="Excludes: "))
					
					self.add_event(name, day, day, True)
				continue

			#
			if line[5]:
				misc_parts = line[5].split('|')
				i = 0

				while(i < len(misc_parts)-1):
					#
					if(misc_parts[i] is 't'):
						if misc_parts[i+1] == '1':
							days = True
						elif misc_parts[i+1] == '2':
							weeks = True
						elif misc_parts[i+1] == '5':
							months = True
					elif(misc_parts[i] is 'i'):
						repeats_every = int(misc_parts[i+1])
					elif(misc_parts[i] is 's'):
						limit_start = datetime.datetime.fromtimestamp(int(misc_parts[i+1])/1000)
					elif(misc_parts[i] is 'e'):
						limit_end =  datetime.datetime.fromtimestamp(int(misc_parts[i+1])/1000)
					elif(misc_parts[i] is 'a'):
						a = 1
					i+=2
				#/while
			

				# Exceptions 1
				if(line[6]):
					excludes_tmp = line[6].split(',')
					i = 0
					while(i < len(excludes_tmp)):
						excludes_tmp[i] = datetime.datetime.fromtimestamp(int(excludes_tmp[i])/1000)
						i+=1
					excludes += excludes_tmp
				#/if
				# Exceptions 2
				if(line[7]):
					excludes_tmp = line[7].split(',')
					i = 0
					while(i < len(excludes_tmp)):
						excludes_tmp[i] = datetime.datetime.fromtimestamp(int(excludes_tmp[i])/1000)
						i+=1
					excludes += excludes_tmp
				#/if

				#
				if(limit_end and _end_limit.date() < limit_end.date()):
					limit_end = _end_limit
				else:
					if (_end_limit.date() < self.limit.date()):
						limit_end = _end_limit
					else:
						limit_end = self.limit
				#/if

				# Set '_end' to the correct day
				_start, _end = self.sanitize_dates(_start, _end)
				_hour = _end.hour
				_minute = _end.minute

				if ((limit_end.date() - _start.date()).days > 0):

					# Skip excluded days in repeating series
					if(not self.date_in_excludes(_start, excludes)):
						if(DEBUG >= 2): 
							print("rep1", self.dates_to_str(_start, prefix="Start: "), self.dates_to_str(_end, prefix="End: "), self.dates_to_str(limit_end, allday=True, prefix="Limit: "), name, self.dates_to_str(excludes, allday=True, prefix="Excludes:"))
						self.add_event(name, _start, _end, _allDay)
					elif(DEBUG >= 2): 
						print("rep1", "EXCLUDED", self.dates_to_str(_start, prefix="Start: "), self.dates_to_str(_end, prefix="End: "), self.dates_to_str(limit_end, allday=True, prefix="Limit: "), name, self.dates_to_str(excludes, allday=True, prefix="Excludes:"))
				else:
					# Shouldn't happen
					if(DEBUG >= 2): 
						print("rep1", "ERROR", self.dates_to_str(_start, prefix="Start: "), self.dates_to_str(_end, prefix="End: "), self.dates_to_str(limit_end, allday=True, prefix="Limit: "), name, self.dates_to_str(excludes, allday=True, prefix="Excludes:"))
					self.add_event(name, _start, _end, _allDay)
				
				#
				while (limit_end.date() - _start.date()).days > 0:
					#
					if days:
						_start = _start + datetime.timedelta(days=repeats_every)
					if weeks:
						_start = _start + datetime.timedelta(days=repeats_every*7)
					if months:
						_start = _start + datetime.timedelta(days=repeats_every*365/12)
					_end = _start.replace(hour=_hour, minute=_minute)

					# Skip excluded days in repeating series
					if (not self.date_in_excludes(_start, excludes)):
						if(DEBUG >= 2): 
							print("rep2", self.dates_to_str(_start, prefix="Start: "), self.dates_to_str(_end, prefix="End: "), name, self.dates_to_str(excludes, allday=True, prefix="Excludes:"))
						self.add_event(name, _start, _end, _allDay)
					elif(DEBUG >= 2): 
						print("rep2", "EXCLUDED", self.dates_to_str(_start, prefix="Start: "), self.dates_to_str(_end, prefix="End: "), name, self.dates_to_str(excludes, allday=True, prefix="Excludes:"))
				#/while
				counter+=1
			else:
				_start, _end = self.sanitize_dates(_start, _end)
				if(DEBUG >= 2): 
					print("single", self.dates_to_str(_start, prefix="Start: "), self.dates_to_str(_end, prefix="End: "), name, self.dates_to_str(excludes, allday=True, prefix="Excludes:"))
				self.add_event(name, _start, _end, _allDay)
			#/if
			if(DEBUG >= 2): 
				print("")
				print("")
		#/for

	##
	def dates_to_str(self, dates, allday=False, prefix=""):
		result = ""
		if(type(dates) is datetime.datetime or type(dates) is datetime.date):
			if(allday): result = prefix + dates.strftime("%Y-%m-%d")
			else: 	    result = prefix + dates.strftime("%Y-%m-%d %H:%Mz")
			return result

		result = []
		if(prefix):
			result.append(prefix)
		for date in dates:
			if(allday): result.append(date.strftime("%Y-%m-%d"))
			else: 	    result.append(date.strftime("%Y-%m-%d %H:%Mz"))
		return result

	##
	def date_in_excludes(self, evDate, excludes):
		for ex in excludes:
			if (ex.date() - evDate.date() == datetime.timedelta(0)):
				return True
		return False

	## 
	def sanitize_dates(self, start, end):
		if (end-start).days > 0:
			end = end.replace(year=start.year, month=start.month, day=start.day)
		return (start, end)

	## 
	def add_event(self, name, start, end, allDay, location=''):
		event = Event(name=name, begin=start, end=end, location=location)
		
		if allDay:
			event.make_all_day()

		if (start-self.lower_limit).days > 0:
			self.calendar.events.append(event)

	## 
	def save_to_file(self, filename):
		with open(filename, 'w') as cal_file:
			cal_file.writelines(self.calendar)

	## 
	def generate_ical_file(self, filename):
		global DEBUG
		
		if(DEBUG): 
			print("Generating the calendar...")
		self.build_ical_structure()

		if(DEBUG): 
			print("Exporting the result into file " + filename + "...")
		self.save_to_file(filename)

	## 
	def generate_calendar_sync_to_ox(self):
		global DEBUG

		if(DEBUG): 
			print("Generating the calendar...")
		self.build_ical_structure()

		# 'needs_normalization' has to be set to false. Else it will try to normalize the times of the events to GMT, which are already at Zulu
		if(DEBUG): 
			print("Syncing the result with the OX calendar...")
		self.dav.parse_and_upload_data(self.calendar, needs_normalization=False)

## Check, if all needed files containing the credentials exist
def check_needed_files():
	if not os.path.isfile(CALDAV_CREDENTIALS_FILE):
		raise Exception("%s not found, exiting." % (CALDAV_CREDENTIALS_FILE,))
	if not os.path.isfile(OX_CONFIG_FILE):
		raise Exception("%s not found, exiting." % (OX_CONFIG_FILE,))

## Read the files containing the credentials, parse and return them
def import_config():
	config = ConfigParser.ConfigParser()
	config.readfp(open(CALDAV_CREDENTIALS_FILE))
	config.readfp(open(OX_CONFIG_FILE))
	return config

## MAIN ##
def main():
	global DEBUG, DEBUG_TMP
	DEBUG_TMP = True

	# Arguments
	parser = optparse.OptionParser()
	parser.add_option("-d", "--debuglevel", dest="debuglevel", help="Debug output", type="int", action="store", default=0)
	parser.add_option("-f", "--file", dest="filename", help="Export result into file instead of OX", type="string", action="store", default=None)
	opts, args = parser.parse_args()
	
	# Debug Output
	if opts.debuglevel:
		DEBUG = opts.debuglevel

	# Check if credential Files exist and import them
	check_needed_files()
	config = import_config()
	
	# Build up a SSH tunnel onto the OX server to reach the SQL database
	if(DEBUG): 
		print("Building up a tunnel to the OX server...")
	tunnel = Tunnel(config)
	tunnel.connect()

	# Query the needed data from the OX database
	if(DEBUG): 
		print("Querying the SQL database...")
	if tunnel.port:
		mysql = Mysql(tunnel.port, config)
		query_result = mysql.do_query()
	tunnel.disconnect()

	# Generate the calender in iCalendar format and export it into the OX ressource calendar
	# (OR export it into a file)
	if query_result:
		ical = Ical(query_result, config)

		if(opts.filename):
			ical.generate_ical_file(opts.filename)
		else:
			ical.generate_calendar_sync_to_ox()
		if(DEBUG): 
			print("Done.")
	else:
		print("Received an empty answer from the SQL database. Quitting.")

## RUN ##
if __name__ == "__main__":
	main()
